import logging

from datapizza.core.clients import Client
from datapizza.core.modules.rewriter import Rewriter
from datapizza.memory.memory import Memory
from datapizza.tools import Tool
from datapizza.tools.tools import tool
from datapizza.type.type import FunctionCallBlock, TextBlock

log = logging.getLogger(__name__)


class ToolRewriter(Rewriter):
    """
    A tool-based query rewriter that uses LLMs to transform user queries through structured tool interactions.
    """

    def __init__(
        self,
        client: Client,
        system_prompt: str | None = None,
        tool: Tool | None = None,
        tool_choice: str = "required",
        tool_output_name: str = "query",
        **invoke_args,
    ):
        self.client = client
        self.system_prompt = system_prompt
        self.invoke_args = invoke_args or {}

        if tool is None:
            # using a default tool
            self.tool = Tool(
                name="query_database",
                description="Query a database",
                func=self._search_vectorstore,
            )
        else:
            self.tool = tool

        self.tool_choice = tool_choice
        self.tool_output_name = tool_output_name

    def rewrite(self, user_prompt: str, memory: Memory | None = None) -> str:
        """
        Args:
            user_prompt: The user query to rewrite.
            memory: The memory to use for the rewrite.
            kwargs: Additional arguments to pass to the client.

        Returns:
            The rewritten query.
        """
        response = self.client.invoke(
            input=user_prompt,
            system_prompt=self.system_prompt,
            memory=memory,
            tool_choice=self.tool_choice,
            tools=[self.tool],
            **self.invoke_args,
        )

        if len(response.content) != 1:
            raise ValueError(
                "ToolRewriter supposed to return only one response, something bad occured"
            )

        content = response.content[0]

        if isinstance(content, FunctionCallBlock):
            return content.arguments[self.tool_output_name]
        elif isinstance(content, TextBlock):
            log.warning(
                "LLM returned TextBlock instead of FunctionCallBlock. "
                "The model may not support tool_choice='required' or the query was too generic. "
                "Returning original user prompt as fallback."
            )
            return user_prompt
        else:
            raise ValueError(
                f"ToolRewriter expected FunctionCallBlock or TextBlock, got {type(content).__name__}"
            )

    async def a_rewrite(self, user_prompt: str, memory: Memory | None = None) -> str:
        """
        Args:
            user_prompt: The user query to rewrite.
            memory: The memory to use for the rewrite.
            kwargs: Additional arguments to pass to the client.

        Returns:
            The rewritten query.
        """
        response = await self.client.a_invoke(
            input=user_prompt,
            system_prompt=self.system_prompt,
            memory=memory,
            tool_choice=self.tool_choice,
            tools=[self.tool],
            **self.invoke_args,
        )

        if len(response.content) != 1:
            raise ValueError(
                "ToolRewriter supposed to return only one response, something bad occured"
            )

        content = response.content[0]

        if isinstance(content, FunctionCallBlock):
            return content.arguments[self.tool_output_name]
        elif isinstance(content, TextBlock):
            log.warning(
                "LLM returned TextBlock instead of FunctionCallBlock. "
                "The model may not support tool_choice='required' or the query was too generic. "
                "Returning original user prompt as fallback."
            )
            return user_prompt
        else:
            raise ValueError(
                f"ToolRewriter expected FunctionCallBlock or TextBlock, got {type(content).__name__}"
            )

    @tool
    def _search_vectorstore(self, query: str):
        """
        Search the vectorstore for the most relevant chunks

        Args:
            query: The query to search the vectorstore for

        Returns:
            A list of Chunks that are most relevant to the query
        """
        pass



from pathlib import Path

from datapizza.core.models import PipelineComponent
from datapizza.pipeline import IngestionPipeline
from datapizza.type import Chunk, DenseEmbedding, Node

# Get the directory containing this test file for relative path resolution
TEST_DIR = Path(__file__).parent


class CustomSplitter(PipelineComponent):
    def __init__(self):
        pass

    def _run(self, nodes: list[Node]):
        """Takes a list of Nodes and returns them (dummy implementation)."""
        if not isinstance(nodes, list):
            raise TypeError(f"Expected input to be a list of Nodes, got {type(nodes)}")
        return nodes

    async def _a_run(self, nodes: list[Node]):
        """Takes a list of Nodes and returns them (dummy implementation)."""
        if not isinstance(nodes, list):
            raise TypeError(f"Expected input to be a list of Nodes, got {type(nodes)}")
        return nodes


class SplitterWrapper(PipelineComponent):
    """A wrapper component that accepts a splitter element for testing elements feature."""

    def __init__(self, splitter: PipelineComponent):
        self.splitter = splitter

    def _run(self, data):
        """Delegates to the wrapped splitter."""
        return self.splitter(data)

    async def _a_run(self, data):
        """Delegates to the wrapped splitter asynchronously."""
        return await self.splitter.a_run(data)


def test_pipeline():
    pipeline = IngestionPipeline(
        modules=[
            CustomSplitter(),
        ]
    )
    assert len(pipeline.pipeline.components) == 1
    assert isinstance(pipeline.pipeline.components[0], CustomSplitter)


def test_pipeline_from_yaml():
    from datapizza.modules.captioners.llm_captioner import LLMCaptioner
    from datapizza.modules.splitters.node_splitter import NodeSplitter

    pipeline = IngestionPipeline().from_yaml(str(TEST_DIR / "config.yaml"))
    assert isinstance(pipeline.pipeline.components[1], LLMCaptioner)
    assert isinstance(pipeline.pipeline.components[2], NodeSplitter)
    assert pipeline.collection_name == "test"


def test_pipeline_from_yaml_with_constants():
    pipeline = IngestionPipeline().from_yaml(str(TEST_DIR / "config.yaml"))
    assert (
        pipeline.pipeline.components[1].system_prompt_table
        == "You are a helpful assistant that captions tables."
    )
    assert (
        pipeline.pipeline.components[1].system_prompt_figure
        == "You are a helpful assistant that captions figures."
    )


def test_ingestion_pipeline():
    from datapizza.modules.splitters.text_splitter import TextSplitter

    class FakeEmbedder(PipelineComponent):
        def _run(self, nodes: list[Chunk]):
            for node in nodes:
                node.embeddings = [
                    DenseEmbedding(name="embedding", vector=[0.0] * 1536)
                ]
            return nodes

        async def _a_run(self, nodes: list[Chunk]):
            return self._run(nodes)

    pipeline = IngestionPipeline(
        modules=[
            TextSplitter(max_char=300),
            FakeEmbedder(),
        ],
        collection_name="test",
    )

    chunks = pipeline.run("Ciao, questo è del testo da ingestionare")

    assert len(chunks) > 0


def test_ingestion_pipeline_with_multiple_files():
    """Test issue #63: IngestionPipeline should handle list of file paths."""
    from datapizza.modules.splitters.text_splitter import TextSplitter

    class FakeEmbedder(PipelineComponent):
        def _run(self, nodes: list[Chunk]):
            for node in nodes:
                node.embeddings = [
                    DenseEmbedding(name="embedding", vector=[0.0] * 1536)
                ]
            return nodes

    pipeline = IngestionPipeline(
        modules=[
            TextSplitter(max_char=300),
            FakeEmbedder(),
        ],
        collection_name="test_multi",
    )

    # Test with list of strings
    chunks = pipeline.run(
        ["Primo documento", "Secondo documento", "Terzo documento"],
        metadata={"batch": "test"},
    )

    assert len(chunks) == 3, f"Expected 3 chunks, got {len(chunks)}"


def test_ingestion_pipeline_list_validation():
    """Test that invalid list elements are rejected."""
    from datapizza.modules.splitters.text_splitter import TextSplitter

    pipeline = IngestionPipeline(
        modules=[TextSplitter(max_char=300)],
    )

    # Should raise ValueError for non-string elements
    try:
        pipeline.run([123, "valid_string", None])
        raise AssertionError("Should have raised ValueError")
    except ValueError as e:
        assert "must be strings" in str(e).lower()


def test_ingestion_pipeline_empty_list():
    """Test that empty list is handled gracefully."""
    from datapizza.modules.splitters.text_splitter import TextSplitter

    class FakeEmbedder(PipelineComponent):
        def _run(self, nodes: list[Chunk]):
            for node in nodes:
                node.embeddings = [
                    DenseEmbedding(name="embedding", vector=[0.0] * 1536)
                ]
            return nodes

    pipeline = IngestionPipeline(
        modules=[
            TextSplitter(max_char=300),
            FakeEmbedder(),
        ],
        collection_name="test_empty",
    )

    # Should handle empty list without errors
    chunks = pipeline.run([])

    assert len(chunks) == 0, "Should have 0 chunks for empty list"


def test_pipeline_from_yaml_with_elements():
    """Test that elements section in YAML config correctly instantiates and injects components."""
    pipeline = IngestionPipeline().from_yaml(
        str(TEST_DIR / "config_with_elements.yaml")
    )

    # Verify the pipeline has the expected components
    assert len(pipeline.pipeline.components) == 3

    # The second component should be SplitterWrapper with a TextSplitter element
    splitter_wrapper = pipeline.pipeline.components[1]
    # Use class name check because importlib loads from a different module context
    assert splitter_wrapper.__class__.__name__ == "SplitterWrapper"
    assert splitter_wrapper.splitter.__class__.__name__ == "TextSplitter"
    # Verify the element was instantiated with the correct params
    assert splitter_wrapper.splitter.max_char == 2000

    # Verify collection name
    assert pipeline.collection_name == "test_elements"


def test_pipeline_from_yaml_elements_with_constants():
    """Test that elements and constants can coexist and both work correctly."""
    pipeline = IngestionPipeline().from_yaml(
        str(TEST_DIR / "config_with_elements.yaml")
    )

    # Verify constants still work (from the third component - LLMCaptioner)
    captioner = pipeline.pipeline.components[2]
    assert (
        captioner.system_prompt_table
        == "You are a helpful assistant that captions tables."
    )
    assert (
        captioner.system_prompt_figure
        == "You are a helpful assistant that captions figures."
    )


def test_pipeline_from_yaml_backward_compatibility():
    """Test that existing YAML configs without elements section still work."""
    # This uses the original config.yaml which doesn't have elements section
    pipeline = IngestionPipeline().from_yaml(str(TEST_DIR / "config.yaml"))

    # Should work exactly as before
    assert len(pipeline.pipeline.components) == 4
    assert pipeline.collection_name == "test"



import os

from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.trace import ProxyTracerProvider


class MissingDatapizzaConfigurationError(ValueError):
    pass


class DatapizzaMonitoringInstrumentor:
    def __init__(
        self,
        *,
        api_key: str | None = None,
        project_id: str | None = None,
        service_name: str | None = None,
        service_version: str | None = None,
        endpoint: str | None = None,
    ):
        resolved_api_key = self._resolve_optional_value(api_key, "DATAPIZZA_API_KEY")
        resolved_project_id = self._resolve_optional_value(
            project_id, "DATAPIZZA_PROJECT_ID"
        )

        missing = []
        if resolved_api_key is None:
            missing.append("DATAPIZZA_API_KEY")
        if resolved_project_id is None:
            missing.append("DATAPIZZA_PROJECT_ID")
        resolved_endpoint = self._resolve_optional_value(
            endpoint, "DATAPIZZA_OTLP_ENDPOINT"
        )
        if resolved_endpoint is None:
            missing.append("DATAPIZZA_OTLP_ENDPOINT")
        if missing:
            missing_list = ", ".join(missing)
            raise MissingDatapizzaConfigurationError(
                f"Missing required configuration: {missing_list}"
            )

        self.api_key = resolved_api_key
        self.project_id = resolved_project_id
        self.service_name = (
            self._resolve_optional_value(service_name, "OTEL_SERVICE_NAME")
            or "datapizza"
        )
        self.service_version = (
            self._resolve_optional_value(service_version, "OTEL_SERVICE_VERSION")
            or "0.1.0"
        )
        self.endpoint = resolved_endpoint

        self._is_instrumented = False

    @classmethod
    def from_env(cls):
        return cls()

    @staticmethod
    def _clean_value(value: str | None) -> str | None:
        if value is None:
            return None

        cleaned = value.strip()
        if not cleaned:
            return None

        return cleaned

    def _resolve_optional_value(
        self, explicit_value: str | None, env_name: str
    ) -> str | None:
        value = self._clean_value(explicit_value)
        if value is not None:
            return value

        return self._clean_value(os.getenv(env_name))

    def instrument(self) -> None:
        if self._is_instrumented:
            return

        resource = Resource.create(
            {
                "service.name": self.service_name,
                "service.version": self.service_version,
            }
        )

        exporter = OTLPSpanExporter(
            endpoint=self.endpoint,
            headers={
                "x-project-id": self.project_id,
                "authorization": f"Bearer {self.api_key}",
            },
        )

        current_provider = trace.get_tracer_provider()
        if isinstance(current_provider, ProxyTracerProvider):
            trace.set_tracer_provider(TracerProvider(resource=resource))
            current_provider = trace.get_tracer_provider()

        if not hasattr(current_provider, "add_span_processor"):
            raise RuntimeError(
                "Tracer provider does not support span processors. "
                "Set an OpenTelemetry SDK TracerProvider before calling instrument()."
            )

        span_processor = BatchSpanProcessor(exporter)
        current_provider.add_span_processor(span_processor)
        self._is_instrumented = True

    def get_tracer(self, name: str, version: str | None = None):
        if not self._is_instrumented:
            raise RuntimeError(
                "Datapizza monitoring is not instrumented. Call instrument() first."
            )

        return trace.get_tracer(name, version)



from rich.console import Console

console = Console()

from datapizza.tracing.instrumentor import (
    DatapizzaMonitoringInstrumentor,
    MissingDatapizzaConfigurationError,
)
from datapizza.tracing.tracing import ContextTracing

__all__ = [
    "ContextTracing",
    "DatapizzaMonitoringInstrumentor",
    "MissingDatapizzaConfigurationError",
]



from unittest.mock import Mock, patch

import pytest

from datapizza.tracing.instrumentor import (
    DatapizzaMonitoringInstrumentor,
    MissingDatapizzaConfigurationError,
)


@pytest.fixture
def required_env(monkeypatch):
    monkeypatch.setenv("DATAPIZZA_API_KEY", "test_api_key")
    monkeypatch.setenv("DATAPIZZA_PROJECT_ID", "test_project_id")
    monkeypatch.setenv("DATAPIZZA_OTLP_ENDPOINT", "http://localhost:4318/v1/traces")


def test_instrumentor_reads_required_env(required_env):
    instrumentor = DatapizzaMonitoringInstrumentor()

    assert instrumentor.api_key == "test_api_key"
    assert instrumentor.project_id == "test_project_id"


def test_instrumentor_raises_when_required_env_missing(monkeypatch):
    monkeypatch.delenv("DATAPIZZA_API_KEY", raising=False)
    monkeypatch.delenv("DATAPIZZA_PROJECT_ID", raising=False)
    monkeypatch.delenv("DATAPIZZA_OTLP_ENDPOINT", raising=False)

    with pytest.raises(MissingDatapizzaConfigurationError):
        DatapizzaMonitoringInstrumentor()


def test_explicit_values_override_env(required_env):
    instrumentor = DatapizzaMonitoringInstrumentor(
        api_key="explicit_api_key",
        project_id="explicit_project_id",
        service_name="test-service",
        service_version="9.9.9",
        endpoint="http://localhost:4318/v1/traces",
    )

    assert instrumentor.api_key == "explicit_api_key"
    assert instrumentor.project_id == "explicit_project_id"
    assert instrumentor.service_name == "test-service"
    assert instrumentor.service_version == "9.9.9"
    assert instrumentor.endpoint == "http://localhost:4318/v1/traces"


def test_get_tracer_requires_instrument(required_env):
    instrumentor = DatapizzaMonitoringInstrumentor()

    with pytest.raises(RuntimeError, match=r"Call instrument\(\) first"):
        instrumentor.get_tracer(__name__)


def test_instrument_adds_span_processor_once(required_env):
    fake_provider = Mock()

    with (
        patch(
            "datapizza.tracing.instrumentor.trace.get_tracer_provider",
            return_value=fake_provider,
        ),
        patch("datapizza.tracing.instrumentor.OTLPSpanExporter") as mock_exporter,
        patch("datapizza.tracing.instrumentor.BatchSpanProcessor") as mock_processor,
    ):
        mock_exporter.return_value = Mock()
        mock_processor.return_value = Mock()

        instrumentor = DatapizzaMonitoringInstrumentor()
        instrumentor.instrument()
        instrumentor.instrument()

    assert fake_provider.add_span_processor.call_count == 1


def test_from_env_returns_ready_instrumentor(required_env):
    instrumentor = DatapizzaMonitoringInstrumentor.from_env()

    assert isinstance(instrumentor, DatapizzaMonitoringInstrumentor)
    assert instrumentor.api_key == "test_api_key"



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import AsyncExitStack, asynccontextmanager
from datetime import timedelta
from typing import Any

import mcp.types as types
from mcp.client.session import ClientSession, SamplingFnT
from mcp.client.stdio import StdioServerParameters, stdio_client
from mcp.client.streamable_http import streamablehttp_client
from mcp.shared.session import ProgressFnT
from pydantic import AnyUrl

from datapizza.core.executors.async_executor import AsyncExecutor
from datapizza.tools import Tool


class MCPClient:
    """
    Helper for interacting with Model Context Protocol servers.

    Can be used in two modes:

    1. **Stateless mode** (default): Each operation creates a new session.
       Good for HTTP-based MCP servers that don't require persistence.

       ```python
       client = MCPClient(url="https://example.com/mcp")
       tools = client.list_tools()
       # Each tool call creates a new session
       ```

    2. **Persistent mode**: Use as an async context manager to keep the session
       alive across multiple operations. Required for stdio-based servers or
       servers that maintain state.

       ```python
       async with MCPClient(command="uvx", args=["my-mcp-server"]) as client:
           tools = await client.a_list_tools()
           # All tool calls share the same session
           agent = Agent(tools=tools)
           await agent.a_run("do something")
       ```

    Args:
        url: The URL of the MCP server.
        command: The command to run the MCP server.
        headers: The headers to pass to the MCP server.
        args: The arguments to pass to the MCP server.
        env: The environment variables to pass to the MCP server.
        timeout: The timeout for the MCP server.
        sampling_callback: The sampling callback to pass to the MCP server.
    """

    def __init__(
        self,
        *,
        url: str | None = None,
        command: str | None = None,
        headers: dict[str, str] | None = None,
        args: list[str] | None = None,
        env: dict[str, str] | None = None,
        timeout: int = 30,
        sampling_callback: SamplingFnT | None = None,
    ) -> None:
        self.url = url
        self.command = command
        self.args = args or []
        self.env = env or {}
        self.timeout = timeout
        self.headers = headers or {}
        self.sampling_callback = sampling_callback

        # Persistent session state
        self._persistent_session: ClientSession | None = None
        self._exit_stack: AsyncExitStack | None = None

        if not url and not command:
            raise ValueError("Either url or command must be provided")
        if url and command:
            raise ValueError("Only one of url or command must be provided")

    async def __aenter__(self) -> MCPClient:
        """Enter persistent session mode."""
        self._exit_stack = AsyncExitStack()
        await self._exit_stack.__aenter__()

        if self.url:
            read_stream, write_stream, _ = await self._exit_stack.enter_async_context(
                streamablehttp_client(
                    self.url,
                    headers=self.headers or None,
                    timeout=self._get_timeout,
                )
            )
        elif self.command:
            server_parameters = StdioServerParameters(
                command=self.command,
                args=self.args,
                env=self.env or None,
            )
            read_stream, write_stream = await self._exit_stack.enter_async_context(
                stdio_client(server_parameters)
            )
        else:
            raise ValueError("Either url or command must be provided")

        session = await self._exit_stack.enter_async_context(
            ClientSession(
                read_stream,
                write_stream,
                read_timeout_seconds=self._get_timeout,
                sampling_callback=self.sampling_callback,
            )
        )
        await session.initialize()
        self._persistent_session = session
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        """Exit persistent session mode and cleanup resources."""
        self._persistent_session = None
        if self._exit_stack:
            await self._exit_stack.__aexit__(exc_type, exc_val, exc_tb)
            self._exit_stack = None

    @property
    def _get_timeout(self) -> timedelta:
        return timedelta(seconds=self.timeout)

    @property
    def is_persistent(self) -> bool:
        """Return True if the client is in persistent session mode."""
        return self._persistent_session is not None

    @asynccontextmanager
    async def _session(self) -> AsyncIterator[ClientSession]:
        """
        Yield an initialized ClientSession.

        If the client is in persistent mode (used as async context manager),
        yields the persistent session. Otherwise, creates a new session
        for this single operation.
        """
        # Use persistent session if available
        if self._persistent_session is not None:
            yield self._persistent_session
            return

        # Otherwise create a new session for this operation
        if self.url:
            async with (
                streamablehttp_client(
                    self.url,
                    headers=self.headers or None,
                    timeout=self._get_timeout,
                ) as (read_stream, write_stream, _),
                ClientSession(
                    read_stream,
                    write_stream,
                    read_timeout_seconds=self._get_timeout,
                    sampling_callback=self.sampling_callback,
                ) as session,
            ):
                await session.initialize()
                yield session

        elif self.command:
            server_parameters = StdioServerParameters(
                command=self.command,
                args=self.args,
                env=self.env or None,
            )
            async with (
                stdio_client(server_parameters) as (read_stream, write_stream),
                ClientSession(
                    read_stream,
                    write_stream,
                    read_timeout_seconds=self._get_timeout,
                    sampling_callback=self.sampling_callback,
                ) as session,
            ):
                await session.initialize()
                yield session

    async def call_tool(
        self,
        tool_name: str,
        arguments: dict[str, Any] | None = None,
        progress_callback: ProgressFnT | None = None,
    ) -> types.CallToolResult:
        """
        Call a tool on the MCP server.

        Args:
            tool_name: The name of the tool to call.
            arguments: The arguments to pass to the tool.
            progress_callback: The progress callback to pass to the tool.

        Returns:
            The result of the tool call.
        """
        async with self._session() as session:
            return await session.call_tool(
                tool_name,
                arguments=arguments or {},
                progress_callback=progress_callback,
            )

    def list_tools(self) -> list[Tool]:
        """
        List the tools available on the MCP server.

        Returns:
            A list of :class:`Tool` objects.
        """
        return AsyncExecutor.get_instance().run(self.a_list_tools(), timeout=10)

    async def a_list_tools(self) -> list[Tool]:
        """
        List the tools available on the MCP server.

        Returns:
            A list of :class:`Tool` objects.
        """
        async with self._session() as session:
            result = await session.list_tools()

        tools: list[Tool] = []
        for mcp_tool in result.tools:
            t_name = mcp_tool.name

            def make_execute_tool(tool_name: str):
                async def execute_tool(**kwargs):
                    result = await self.call_tool(
                        tool_name=tool_name,
                        arguments=kwargs or {},
                    )
                    return result.model_dump_json()

                return execute_tool

            schema = mcp_tool.inputSchema or {}
            properties = schema.get("properties", {})
            required = schema.get("required", [])
            tools.append(
                Tool(
                    func=make_execute_tool(t_name),
                    name=mcp_tool.name,
                    description=mcp_tool.description,
                    properties=properties,
                    required=required,
                    strict=bool(schema.get("additionalProperties") is False),
                )
            )

        return tools

    async def list_resources(self) -> types.ListResourcesResult:
        """
        List the resources available on the MCP server.

        Returns:
            A :class:`types.ListResourcesResult` object.
        """
        async with self._session() as session:
            return await session.list_resources()

    async def list_resource_templates(self) -> types.ListResourceTemplatesResult:
        async with self._session() as session:
            return await session.list_resource_templates()

    async def read_resource(self, resource_uri: AnyUrl) -> types.ReadResourceResult:
        async with self._session() as session:
            return await session.read_resource(resource_uri)

    def list_prompts(self) -> types.ListPromptsResult:
        """
        List the prompts available on the MCP server.

        Returns:
            A :class:`types.ListPromptsResult` object.
        """
        return AsyncExecutor.get_instance().run(self.a_list_prompts(), timeout=10)

    async def a_list_prompts(self) -> types.ListPromptsResult:
        """
        List the prompts available on the MCP server.

        Returns:
            A :class:`types.ListPromptsResult` object.
        """
        async with self._session() as session:
            return await session.list_prompts()

    async def get_prompt(
        self, prompt_name: str, arguments: dict[str, str] | None = None
    ) -> types.GetPromptResult:
        """
        Get a prompt from the MCP server.
        """
        async with self._session() as session:
            return await session.get_prompt(prompt_name, arguments or {})



import importlib
import logging
from typing import Any

import yaml

from datapizza.clients import ClientFactory
from datapizza.core.models import PipelineComponent
from datapizza.core.utils import replace_env_vars
from datapizza.core.vectorstore import Vectorstore
from datapizza.type import Chunk

log = logging.getLogger(__name__)


def _replace_element_refs(value: Any, elements: dict[str, Any]) -> Any:
    """
    Replace element references (${element_name}) with actual element instances.

    Args:
        value: The value to process (can be string, dict, list, or any other type)
        elements: Dictionary mapping element names to their instances

    Returns:
        The value with element references replaced by actual instances
    """
    if isinstance(value, str) and value.startswith("${") and value.endswith("}"):
        element_name = value[2:-1]
        if element_name in elements:
            return elements[element_name]
        # If not found in elements, return as-is (might be a constant or env var)
        return value
    elif isinstance(value, dict):
        return {k: _replace_element_refs(v, elements) for k, v in value.items()}
    elif isinstance(value, list):
        return [_replace_element_refs(item, elements) for item in value]
    else:
        return value


def _instantiate_element(element_config: dict[str, Any]) -> Any:
    """
    Instantiate an element from its configuration.

    Args:
        element_config: Dictionary with 'type', 'module', and optional 'params' keys

    Returns:
        The instantiated element
    """
    module_path = element_config["module"]
    class_name = element_config["type"]
    params = element_config.get("params", {})

    module = importlib.import_module(module_path)
    class_ = getattr(module, class_name)
    return class_(**params)


class Pipeline:
    def __init__(self, components: list[PipelineComponent] | None = None):
        self.components = components or []

    def run(self, input_data=None):
        data = input_data
        for component in self.components:
            log.info(f"Running component {component.__class__.__name__}")
            data = component(data)
        return data

    async def a_run(self, input_data=None):
        data = input_data
        for component in self.components:
            log.info(f"Running component {component.__class__.__name__}")
            data = await component.a_run(data)
        return data


class IngestionPipeline:
    """
    A pipeline for ingesting data into a vector store.
    """

    def __init__(
        self,
        modules: list[PipelineComponent] | None = None,
        vector_store: Vectorstore | None = None,
        collection_name: str | None = None,
    ):
        """
        Initialize the ingestion pipeline.

        Args:
            modules (list[PipelineComponent], optional): List of pipeline components. Defaults to None.
            vector_store (Vectorstore, optional): Vector store to store the ingested data. Defaults to None.
            collection_name (str, optional): Name of the vector store collection to store the ingested data. Defaults to None.
        """
        self.pipeline = Pipeline(modules)
        self.vector_store = vector_store
        self.collection_name = collection_name
        self.components = modules

        if self.vector_store and not self.collection_name:
            raise ValueError("Collection name must be set if vector store is provided")

    def run(
        self, file_path: str | list[str], metadata: dict | None = None
    ) -> list[Chunk] | None:
        """Run the ingestion pipeline.

        Args:
            file_path (str | list[str]): The file path or list of file paths to ingest.
            metadata (dict, optional): Metadata to add to the ingested chunks. Defaults to None.

        Returns:
            list[Chunk] | None: If vector_store is not set, returns all accumulated chunks from all files.
                                If vector_store is set, returns None after storing all chunks.
        """
        # Normalize to list for uniform processing
        if isinstance(file_path, str):
            file_paths = [file_path]
        elif isinstance(file_path, list):
            # Validate that all elements are strings
            if not all(isinstance(fp, str) for fp in file_path):
                raise ValueError("All elements in file_path list must be strings")
            file_paths = file_path
        else:
            raise ValueError("file_path must be a string or a list of strings")

        all_chunks = []

        # Process each file path
        for fp in file_paths:
            data = self.pipeline.run(fp)

            if not self.vector_store:
                # If no vector store, accumulate results
                if isinstance(data, list):
                    all_chunks.extend(data)
                else:
                    all_chunks.append(data)
            else:
                # Validate chunk data immediately
                if not isinstance(data, list) or not all(
                    isinstance(item, Chunk) for item in data
                ):
                    raise ValueError(
                        f"Data returned from pipeline for '{fp}' must be a list of Chunk objects"
                    )
                all_chunks.extend(data)

        if not self.vector_store:
            return all_chunks

        # Adding metadata to all accumulated chunks
        if metadata:
            for chunk in all_chunks:
                chunk.metadata.update(metadata)

        # Add all chunks to vector store at once (only if we have chunks)
        if all_chunks:
            self.vector_store.add(all_chunks, self.collection_name)

    async def a_run(
        self, file_path: str | list[str], metadata: dict | None = None
    ) -> list[Chunk] | None:
        """Run the ingestion pipeline asynchronously.

        Args:
            file_path (str | list[str]): The file path or list of file paths to ingest.
            metadata (dict, optional): Metadata to add to the ingested chunks. Defaults to None.

        Returns:
            list[Chunk] | None: If vector_store is not set, returns all accumulated chunks from all files.
                                If vector_store is set, returns None after storing all chunks.
        """
        # Normalize to list for uniform processing
        if isinstance(file_path, str):
            file_paths = [file_path]
        elif isinstance(file_path, list):
            # Validate that all elements are strings
            if not all(isinstance(fp, str) for fp in file_path):
                raise ValueError("All elements in file_path list must be strings")
            file_paths = file_path
        else:
            raise ValueError("file_path must be a string or a list of strings")

        all_chunks = []

        # Process each file path
        for fp in file_paths:
            data = await self.pipeline.a_run(fp)

            if not self.vector_store:
                # If no vector store, accumulate results
                if isinstance(data, list):
                    all_chunks.extend(data)
                else:
                    all_chunks.append(data)
            else:
                # Validate chunk data immediately
                if not isinstance(data, list) or not all(
                    isinstance(item, Chunk) for item in data
                ):
                    raise ValueError(
                        f"Data returned from pipeline for '{fp}' must be a list of Chunk objects"
                    )
                all_chunks.extend(data)

        if not self.vector_store:
            return all_chunks

        # Adding metadata to all accumulated chunks
        if metadata:
            for chunk in all_chunks:
                chunk.metadata.update(metadata)

        # Add all chunks to vector store at once (only if we have chunks)
        if all_chunks:
            await self.vector_store.a_add(all_chunks, self.collection_name)

    def from_yaml(self, config_path: str) -> "IngestionPipeline":
        """
        Load the ingestion pipeline from a YAML configuration file.

        The YAML configuration supports the following sections:
        - constants: Key-value pairs for string substitution using ${VAR_NAME} syntax
        - elements: Reusable component definitions that can be referenced in modules
        - ingestion_pipeline: The main pipeline configuration with clients, modules, vector_store, and collection_name

        Example elements section:
            elements:
                my_embedder:
                    type: GoogleEmbedder
                    module: datapizza.embedders.google
                    params:
                        max_char: 2000

        Elements can be referenced in module params using ${element_name} syntax:
            modules:
                - name: embedder
                  type: ChunkEmbedder
                  module: datapizza.embedders
                  params:
                      client: "${my_embedder}"

        Args:
            config_path (str): Path to the YAML configuration file.

        Returns:
            IngestionPipeline: The ingestion pipeline instance.
        """
        with open(config_path) as file:
            config = yaml.safe_load(file)

        constants = config.get("constants", {})
        # Use skip_unknown=True to allow element references (like ${my_embedder})
        # to pass through without being treated as missing env vars
        config = replace_env_vars(config, constants, skip_unknown=True)

        # Parse and instantiate elements
        elements = {}
        if "elements" in config:
            for element_name, element_config in config["elements"].items():
                try:
                    elements[element_name] = _instantiate_element(element_config)
                except (ImportError, AttributeError) as e:
                    raise ValueError(
                        f"Could not load element '{element_name}' "
                        f"({element_config.get('type', 'N/A')}): {e!s}"
                    ) from e
                except KeyError as e:
                    raise ValueError(
                        f"Missing required key {e!s} in element configuration: {element_config}"
                    ) from e

        clients = {}
        ingestion_pipeline = config["ingestion_pipeline"]
        if "clients" in ingestion_pipeline:
            for client_name, client_config in ingestion_pipeline["clients"].items():
                provider = client_config.pop("provider")
                client = ClientFactory.create(
                    provider, client_config.get("api_key"), client_config.get("model")
                )
                clients[client_name] = client

        components = []
        if "modules" in ingestion_pipeline:
            for component_config in ingestion_pipeline["modules"]:
                try:
                    module_path = component_config["module"]
                    module = importlib.import_module(module_path)
                    class_ = getattr(module, component_config["type"])

                    params = component_config.get("params", {})

                    # Replace element references in params
                    params = _replace_element_refs(params, elements)

                    if "client" in params:
                        client_name = params["client"]
                        # Only do client lookup if it's still a string (not already replaced by element)
                        if isinstance(client_name, str):
                            if client_name not in clients:
                                raise ValueError(
                                    f"Client '{client_name}' not found in clients configuration"
                                )
                            params["client"] = clients[client_name]

                    component_instance = class_(**params)
                    components.append(component_instance)
                except (ImportError, AttributeError) as e:
                    raise ValueError(
                        f"Could not load component {component_config.get('type', 'N/A')}: {e!s}"
                    ) from e
                except KeyError as e:
                    raise ValueError(
                        f"Missing required key {e!s} in module configuration: {component_config}"
                    ) from e

        vector_store = None
        if "vector_store" in ingestion_pipeline:
            vector_store_config = ingestion_pipeline["vector_store"]
            vector_store_type = vector_store_config["type"]
            vector_store_module = importlib.import_module(vector_store_config["module"])
            vector_store_class = getattr(vector_store_module, vector_store_type)
            vector_store_params = vector_store_config.get("params", {})
            vector_store = vector_store_class(**vector_store_params)
            self.vector_store = vector_store

        collection_name = None
        if "collection_name" in ingestion_pipeline:
            collection_name = ingestion_pipeline["collection_name"]
            self.collection_name = collection_name

        self.components = components
        self.pipeline = Pipeline(components)
        return self



"""Tests for MCPClient persistent session functionality."""

from unittest.mock import AsyncMock, MagicMock

import pytest

from ..mcp_client import MCPClient


class TestMCPClientInit:
    """Test MCPClient initialization."""

    def test_init_with_url(self):
        client = MCPClient(url="https://example.com/mcp")
        assert client.url == "https://example.com/mcp"
        assert client.command is None
        assert client._persistent_session is None
        assert client._exit_stack is None

    def test_init_with_command(self):
        client = MCPClient(command="uvx", args=["my-server"])
        assert client.command == "uvx"
        assert client.args == ["my-server"]
        assert client.url is None
        assert client._persistent_session is None

    def test_init_requires_url_or_command(self):
        with pytest.raises(ValueError, match="Either url or command"):
            MCPClient()

    def test_init_rejects_both_url_and_command(self):
        with pytest.raises(ValueError, match="Only one of url or command"):
            MCPClient(url="https://example.com", command="uvx")


class TestMCPClientPersistence:
    """Test MCPClient persistent session mode."""

    def test_is_persistent_false_by_default(self):
        client = MCPClient(url="https://example.com/mcp")
        assert client.is_persistent is False

    def test_is_persistent_true_when_session_set(self):
        """Test is_persistent returns True when _persistent_session is set."""
        client = MCPClient(url="https://example.com/mcp")
        client._persistent_session = MagicMock()
        assert client.is_persistent is True

    def test_is_persistent_false_when_session_none(self):
        """Test is_persistent returns False when _persistent_session is None."""
        client = MCPClient(url="https://example.com/mcp")
        client._persistent_session = None
        assert client.is_persistent is False


class TestMCPClientSessionLogic:
    """Test the _session method logic for persistent vs stateless mode."""

    @pytest.mark.asyncio
    async def test_session_yields_persistent_session_when_available(self):
        """When _persistent_session is set, _session should yield it directly."""
        client = MCPClient(url="https://example.com/mcp")
        mock_session = MagicMock()
        client._persistent_session = mock_session

        async with client._session() as session:
            assert session is mock_session

    @pytest.mark.asyncio
    async def test_session_yields_same_persistent_session_multiple_times(self):
        """Multiple calls to _session should return the same persistent session."""
        client = MCPClient(url="https://example.com/mcp")
        mock_session = MagicMock()
        client._persistent_session = mock_session

        sessions = []
        for _ in range(3):
            async with client._session() as s:
                sessions.append(s)

        assert all(s is mock_session for s in sessions)
        assert len(sessions) == 3


class TestMCPClientContextManager:
    """Test the async context manager entry/exit behavior."""

    def test_exit_clears_persistent_session(self):
        """__aexit__ should clear the persistent session."""
        client = MCPClient(url="https://example.com/mcp")
        # Simulate having a session
        client._persistent_session = MagicMock()

        # Call the sync part of cleanup
        client._persistent_session = None

        assert client._persistent_session is None
        assert client.is_persistent is False


class TestMCPClientCallTool:
    """Test call_tool method behavior."""

    @pytest.mark.asyncio
    async def test_call_tool_uses_session(self):
        """call_tool should use _session to get a session and call the tool."""
        client = MCPClient(url="https://example.com/mcp")

        mock_result = MagicMock()
        mock_session = AsyncMock()
        mock_session.call_tool = AsyncMock(return_value=mock_result)

        # Set persistent session to avoid actual connection
        client._persistent_session = mock_session

        result = await client.call_tool("test_tool", {"arg": "value"})

        mock_session.call_tool.assert_called_once_with(
            "test_tool",
            arguments={"arg": "value"},
            progress_callback=None,
        )
        assert result is mock_result

    @pytest.mark.asyncio
    async def test_call_tool_with_progress_callback(self):
        """call_tool should pass progress_callback to session.call_tool."""
        client = MCPClient(url="https://example.com/mcp")

        mock_session = AsyncMock()
        mock_session.call_tool = AsyncMock(return_value=MagicMock())
        client._persistent_session = mock_session

        progress_cb = MagicMock()
        await client.call_tool("test_tool", {"arg": "value"}, progress_cb)

        mock_session.call_tool.assert_called_once_with(
            "test_tool",
            arguments={"arg": "value"},
            progress_callback=progress_cb,
        )


class TestMCPClientListTools:
    """Test list_tools related methods."""

    @pytest.mark.asyncio
    async def test_a_list_tools_uses_session(self):
        """a_list_tools should use _session to list tools."""
        client = MCPClient(url="https://example.com/mcp")

        # Create mock tool
        mock_tool = MagicMock()
        mock_tool.name = "test_tool"
        mock_tool.description = "A test tool"
        mock_tool.inputSchema = {
            "properties": {"arg1": {"type": "string"}},
            "required": ["arg1"],
        }

        mock_result = MagicMock()
        mock_result.tools = [mock_tool]

        mock_session = AsyncMock()
        mock_session.list_tools = AsyncMock(return_value=mock_result)
        client._persistent_session = mock_session

        tools = await client.a_list_tools()

        mock_session.list_tools.assert_called_once()
        assert len(tools) == 1
        assert tools[0].name == "test_tool"
        assert tools[0].description == "A test tool"


class TestMCPClientListPrompts:
    """Test list_prompts related methods."""

    @pytest.mark.asyncio
    async def test_a_list_prompts_uses_session(self):
        """a_list_prompts should use _session to list prompts."""
        client = MCPClient(url="https://example.com/mcp")

        mock_result = MagicMock()
        mock_session = AsyncMock()
        mock_session.list_prompts = AsyncMock(return_value=mock_result)
        client._persistent_session = mock_session

        result = await client.a_list_prompts()

        mock_session.list_prompts.assert_called_once()
        assert result is mock_result


class TestMCPClientResources:
    """Test resource-related methods."""

    @pytest.mark.asyncio
    async def test_list_resources_uses_session(self):
        """list_resources should use _session."""
        client = MCPClient(url="https://example.com/mcp")

        mock_result = MagicMock()
        mock_session = AsyncMock()
        mock_session.list_resources = AsyncMock(return_value=mock_result)
        client._persistent_session = mock_session

        result = await client.list_resources()

        mock_session.list_resources.assert_called_once()
        assert result is mock_result

    @pytest.mark.asyncio
    async def test_list_resource_templates_uses_session(self):
        """list_resource_templates should use _session."""
        client = MCPClient(url="https://example.com/mcp")

        mock_result = MagicMock()
        mock_session = AsyncMock()
        mock_session.list_resource_templates = AsyncMock(return_value=mock_result)
        client._persistent_session = mock_session

        result = await client.list_resource_templates()

        mock_session.list_resource_templates.assert_called_once()
        assert result is mock_result
